WIDGET_HOLD_TIME = 16
WIDGET_REPEAT_TIME = 2
WIDGET_CLICK_TIME = 5
WIDGET_CLICK_DIST = 3
WIDGET_PANEL_COLOR = 1
WIDGET_BACKGROUND_COLOR = 7
WIDGET_SHADOW_COLOR = 13

BUTTON_ENABLED_COLOR = 12
BUTTON_DISABLED_COLOR = 5
BUTTON_PRESSED_COLOR = 7
BUTTON_TEXT_COLOR = 1
BUTTON_PRESSING_TIME = 4

INPUT_TEXT_COLOR = 1
INPUT_FIELD_COLOR = 10
