// Input
pub type Key = u32;
pub type KeyValue = i32;

// Graphics
pub type Rgb8 = u32;
pub type Color = u8;
pub type Tile = (u8, u8);

// Audio
pub type Note = i8;
pub type Tone = u8;
pub type Volume = u8;
pub type Effect = u8;
pub type Speed = u32;
